#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['foodx_devops_tools',
 'foodx_devops_tools.azure',
 'foodx_devops_tools.azure.cloud',
 'foodx_devops_tools.deploy_me',
 'foodx_devops_tools.deploy_me.application_steps',
 'foodx_devops_tools.pipeline_config',
 'foodx_devops_tools.puff',
 'foodx_devops_tools.release_flow',
 'foodx_devops_tools.utilities']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles >=0.7.0, <1.0',
 'ansible >=4.4.0, <6.0',
 'azure-devops >=6.0.0b4, <7.0',
 'click >=8.1.2, <9',
 'deepmerge >=1.0.1, <2',
 'jinja2 >=3.1.1, <4',
 'pydantic >=1.9.0, <2',
 'ruamel.yaml >=0.17.20, <1.0']

extras_require = \
{'dev': ['build_harness >=1.0, <2',
         'pre_commit >=2.17.0, <3',
         'pyjson5 >=1.5.2, <2',
         'types-aiofiles >=0.1.9, <1.0'],
 'doc': ['sphinx >=4.5.0, <5', 'sphinx_rtd_theme >=1.0, <2'],
 'test': ['asynctest >=0.13.0, <1.0',
          'pytest >=7.1.1, <8',
          'pytest-asyncio >=0.18.3, <1.0',
          'pytest-cov >=3.0, <4',
          'pytest-mock >=3.7.0, <4']}

entry_points = \
{'console_scripts': ['deploy-me = '
                     'foodx_devops_tools.deploy_me_entry:flit_entry',
                     'file-maintainer = '
                     'foodx_devops_tools.file_maintainer_entry:flit_entry',
                     'foodx-release-flow = '
                     'foodx_devops_tools.release_flow_entry:flit_entry',
                     'puff = foodx_devops_tools.puff_utility:entrypoint',
                     'validate-configuration = '
                     'foodx_devops_tools.validate_configuration:flit_entry']}

setup(name='foodx_devops_tools',
      version='0.15.1.post2',
      description='Foodx DevOps pipeline utilities.',
      author='FoodX Technologies',
      author_email='support@foodxtech.com',
      url='https://github.com/Food-X-Technologies/foodx_devops_tools',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='>=3.8',
     )
